#ifndef CUA_HPP
#define CUA_HPP

#include <utility>
using namespace std;


template <class T> class Cua {

private:

  struct node{
    T   info;
    node *seg;
  };

  int  llarg;
  node *prim;
  node *ult;


public:

  /* crea una cua buida*/
  Cua(){
    prim = NULL;
    ult = NULL;
    llarg = 0;
  }
  
  /* crea una cua copia d'original*/
  Cua(const Cua& o){
    llarg = o.llarg;
    if (o.prim!=NULL)
      prim= copia(o.prim,ult);
    else prim=ult=NULL;
  }

  Cua& operator=(const Cua& o){
    if (this!=&o){
      llarg = o.llarg;
      allibera(prim);
      if (o.prim!=NULL){ 
	prim= copia(o.prim,ult);
      }
      else prim=ult=NULL;
    }
    return *this;
  }
  
  //Destructor
  ~Cua(){
    allibera(prim);
  }

  //Modificadores 
  

  /* afegeix l'element x al final de la cua */
  void demanar_torn(const T &elem){
    node *aux = new node;
    aux->info = elem;
    aux->seg=NULL;
    if (ult==NULL) prim=aux;
    else ult->seg=aux;
    ult=aux;
    ++llarg;
  }

  /* elimina el primer element de la cua,
     si la cua s buida dona un error */
  void avancar(){
    node *daux = prim;
    if (prim->seg==NULL) ult=NULL;
    prim=prim->seg;
    delete daux;
    --llarg;
  }

  //Cosultores 

  /* torna el primer valor afegit a la cua,
     si la cua s buida dona un error */
  T primer() const {
    return prim->info; 
  }
  
  /* ens retorna si la cua s buida o no */
  bool es_buida() const{
    return (prim==NULL);
  }
  
  /* ens retorna la mida de la cua */
  int mida() const{
    return llarg;
  }
  

private:

  //Pre: n!= NULL
  static node* copia(node* n, node* &nult){
    node *aux = new node;
    aux->info=n->info;
    if (!n->seg) {
      aux->seg = NULL;
      nult=aux;
    }
    else aux->seg = copia(n->seg,nult);
    return aux;
  }
  

  static void allibera(node* n){
    if (n) {
      allibera(n->seg);
      delete(n);
    }
  }

};


#endif


/*
pair<node*,node*> copia(node* n){
  if (!n) {
      node* aux = NULL;
      return make_pair(aux,aux);
    }
  else{
    node *aux = new node;
    node *out = aux;
    aux->info=n->info;
    while (n->seg!=NULL) {
      (*aux).seg = new node;
      n=n->seg;
      aux=aux->seg;
      aux->info=n->info;
    }
    aux->seg=NULL;
    return make_pair(out,aux);
  }
}

*/
