#include <iostream>

using namespace std;

#ifndef LLISTA_HPP
#define LLISTA_HPP

template <class T> class Llista {

// Tipus de mòdul : dades
// Descripció del tipus: Estructura lineal que conté elements de tipus T, que es 
// pot començar a consultar pels extrems, on des de cada element es pot accedir
// a l'element anterior i posterior (si existeixen), i que admet afegir-hi 
// i esborrar-hi elements a qualsevol punt

private:

  struct node{
    T   info;
    node *seg;
    node *ant;
  };

  int  longitud;
  node *prim;
  node *ult;
  node *act;


public:

// Constructora

  Llista(){
    act=NULL;
    prim=NULL;
    ult=NULL;
    longitud= 0;  
  }

  Llista(const Llista& orig){
    copia(orig.prim,orig.act, act, prim, ult);
    longitud=orig.longitud;
  }
  
  Llista& operator=(const Llista& o){
    if (this!=&o){
      longitud = o.longitud;
      allibera(prim);
      copia(o.prim, o.act,act,prim,ult);
    }
    return *this;
  }

  // Destructora: Esborra automàticament els objectes locals en sortir
  // d'un àmbit de visibilitat

  ~Llista(){
    allibera(prim);
  }

  // Modificadores 

  //L'actual está al final quan esta en el següent de l'ultim
  
  //act = NULL significa que l'actual és al final
  
  /* Pre: actual referencia algun element existent o 
     esta al final */
  void afegir(const T &x){
    node *aux= new node;
    aux->seg=act;
    aux->info=x;
    if (act==NULL) {
      if (longitud!=0) ult->seg=aux;
      else prim=aux;
      aux->ant=ult;
      ult=aux;
    }
    else {
      if (act!=prim) act->ant->seg=aux;
      else prim =aux;
      aux->ant=act->ant;
      act->ant=aux;
    }
    ++longitud;
  }
  /* Post: El paràmetre implícit és com el paràmetre implícit original amb x 
     davant de l'actual */



  // Pre: l'actual és algun element existent
  void eliminar(){
    if (act==prim) prim=prim->seg;
    if (act==ult)  ult =act->ant;
    if (act->ant!=NULL) act->ant->seg=act->seg;
    if (act->seg!=NULL) act->seg->ant=act->ant;
    node* aux= act;
    act=act->seg;
    delete aux;
    --longitud;
  }
  /* Post: s'ha eliminat l'element actual original; l'actual passa a
     ser el següent de l'original */ 

  /* Pre: cert */
  void concat(Llista& l){
    if (l.longitud!=0) {
      if (longitud==0) prim= l.prim;
      else {
	ult->seg=l.prim;
	l.prim->ant=ult;
      }
      ult=l.ult;
      longitud+=l.longitud;
      l.prim=NULL;
      l.ult=NULL;
      l.act=NULL;
      l.longitud=0;
    }
    //act= prim;
  }
  /* Post:  El paràmetre implícit té els seus elements originals seguits pels 
     de la llista l original; l està buida*/
  
  
  //Modificadores i consultores d'actual


  void primer(){
    act=prim;
  }
  //l'actual és el primer
  
  void ultim(){
    act=ult;
  }
  //l'actual és l'últim

  void final(){
    act=NULL;
  }
  //l'actual és l'últim

  //l'actual no esta al final
  void seguent(){
    act=act->seg;
  }
  //l'actual és el següent a l'actual original

  //l'actual no és el primer i la llista no és buida
  void anterior(){
    if (act==NULL) act=ult;
    else act=act->ant;
  }
  //l'actual és l'anterior a l'actual original
  
  bool es_final(){
    return (act==NULL);
  }
  // retorna si l'actual esta després l'últim 

  bool es_principi(){
    return (act==prim);
  }
  // retorna si l'actual és igual al primer 
  

  // Pre: actual referencia algun element existent 
  T actual(){
    return act->info;
  }
  // Post: retorna l'element actual


  // Pre: actual referencia algun element existent 
  void modificar_actual(const T &x){
    act->info=x;
  }
  // Post: posa x a l'element actual
  

  // Consultores

  /* Pre: cert */
  bool es_buida() const{
    return (prim==NULL);
  }
  /* Post: El resultat indica si el paràmetre implícit te elements o no */
  
  /* Pre: cert */
  int mida() const{
    return longitud;
  }
  /* Post: El resultat és el nombre d'elements del paràmetre implícit */


  // auxiliars de node

  static void copia(node* n, node* act, 
	     node* &nact, node* &nprim, node* &nult){
    if (n==NULL) {
      nact=NULL;
      nprim=NULL;
      nult=NULL;
    }
    else {
      nprim =new node;
      nprim->info=n->info;
      nprim->ant=NULL;
      copia(n->seg,act,nact,nprim->seg,nult);
      if (nprim->seg!=NULL) nprim->seg->ant=nprim;
      else nult=nprim;
      if (n==act) nact=nprim;
    }
  }



  static void allibera(node* n){
    while (n!=NULL){
      node*aux=n;
      n=n->seg;
      delete aux;
    }
  }


public:

 void invertir(){
   node *aux=prim;
   prim=ult;
   ult=aux;
   invertir_node(aux);
 }


private:

 void invertir_node(node* n){
   if (n!=NULL){
     invertir_node(n->seg);
     node* aux=n->seg;
     n->seg=n->ant;
     n->ant=aux;
   }
 }





// static void copia(node* n, node* act, 
// 	   node* &nact, node* &nprim, node* &nult){
//   if (n==NULL) {
//     nact=NULL;
//     nprim=NULL;
//     nult=NULL;
//   }
//   else {
//     node* aux=new node;
//     nprim=aux;
//     aux->ant=NULL;
//     aux->info=n->info;
//     while (n->seg!=NULL){
//       if (n==act) nact=aux;
//       aux->seg= new node;
//       aux->seg->ant=aux;
//       aux=aux->seg;
//       n=n->seg;
//       aux->info=n->info;
//     }
//     aux->seg=NULL;
//     nult=aux;
//   }
// }

};

#endif

